(* Wolfram Language package *)

(* Implements Derivation and Series Expansion of Operators *)

BeginPackage["QuantumAlgebra`Derivation`", { "QuantumAlgebra`", "QuantumAlgebra`Options`", "QuantumAlgebra`Analyze`" }]


QAD::usage = "Derives an Operator expression depending on some parameter(s)."

QASeries::usage = "Performs a series expansion of an Operator expression."

QASeriesData::usage = "The format returned by QASeries."

LieGroup::usage = "Generates a Lie Group depending on some parameter(s)."

SeriesLieGroup::usage = "Generates a Lie Group and performs a series expansion on it."

Params::usage = "Generates a list of subscripted parameters."

LinearForm::usage = "Builds a Linear Form of generators and subscripted parameters."

GradLinearForm::usage = "Builds a Gradient of a given linear form."

SeriesCoefficients::usage = "Returns a list of Coefficients of a series expansion."

FactorCombinations::usage = "Combines the factors of a linear form to a given order."

DerivCombinations::usage = "Combines the derivations of an operator to a given order."

SeriesFunctionLinearForm::usage = "Returns a series of a function of a given linear form."


Begin["`Private`"]

protected = Unprotect[MakeBoxes, Normal, NonCommutativeMultiply]

(* Test auf "einfachen" Operator *)
SimpleOperatorQ[op_] :=
    MatchQ[op, QuantumAlgebra`Operator[__]] || MatchQ[op, QuantumAlgebra`Operator[__][__]]

SetAttributes[QAD, {HoldFirst}]

(* Ableitungen von "Konstanten" *)
QAD[c_?QuantumAlgebra`ConstantQ, var_] :=
    D[c, var]

(* Ableitungen von einfachen Operatoren *)
QAD[op_?SimpleOperatorQ, var_] :=
    Module[ {res},
        res = D[op, var];
        If[res === 0,
            QuantumAlgebra`Operator[0],
            res
        ]
    ]

(* Ableitung eines Produktes von Konstanter und Operator *)
QAD[Times[c_?QuantumAlgebra`ConstantQ, op_?QuantumAlgebra`OperatorQ], var_] :=
    QAD[c, var] op + c QAD[op, var]

(* Ableitung eines Produktes von 2 Operatoren *)
QAD[NonCommutativeMultiply[op1_?QuantumAlgebra`OperatorQ, op2_?QuantumAlgebra`OperatorQ], var_] :=
    QAD[op1, var] ** op2 + op1 ** QAD[op2, var]

(* Ableitung einer Summe von 2 Operatoren *)
QAD[Plus[op1_?QuantumAlgebra`OperatorQ, op2_?QuantumAlgebra`OperatorQ], var_] :=
    QAD[op1, var] + QAD[op2, var]

(* Ableitung einer ganzzahligen Potenz eines Operators (rekursiv) *)
QAD[Power[op_?QuantumAlgebra`OperatorQ, ex_Integer], var_] /;
	ex >= 2 :=
    QAD[Evaluate[Power[op, ex - 1] ** op], var]

(* Mehrfachableitung eines Operators *)
QAD[op_?QuantumAlgebra`OperatorQ, vars_List, derivs_List] :=
    Fold[QAD[#1, vars[[#2]]]&, op, derivs]


(* The Data Type returned by SeriesFunctionLinearform / SeriesLieGroup *)
QASeriesData[] :=
    0

(* Extrahiert aus einer QASeriesData Liste eine Liste der einzelnen Glieder, die dann addiert werden koennen *)
SeriesItems[qasd:QASeriesData[vars_List, locs_List, factors_List, order_Integer]] :=
    Module[ {AddToAssoc, assoc = <||>, cntvar, k, ccb, op, varex},
    	AddToAssoc[key_, val_] := assoc[key] = (Lookup[assoc, key, QuantumAlgebra`Operator[0]] + val) // QuantumAlgebra`QAExpandAll;
        cntvar = Length[vars];

		For[k = 0, k <= order, k = k + 1,
			For[ccb = 0, ccb < cntvar^k, ccb = ccb + 1,
				op = factors[[k + 1, ccb + 1]];
				varex = Fold[Times[#1, vars[[#2]]]&, 1, Reverse[IntegerDigits[ccb, cntvar, k] + 1]];
				AddToAssoc[varex, varex * op];
			];
		];
        
        Values[assoc]
    ]

(* Ermoeglicht den einfachen Aufruf von MakeBoxes *)
MakeBoxesHelper[exp_] :=
    Module[ {res},
        ClearAttributes[MakeBoxes, {HoldAllComplete}];
        SetAttributes[MakeBoxes, {HoldAll}];
        res = MakeBoxes[Evaluate[exp]];
        ClearAttributes[MakeBoxes, {HoldAll}];
        SetAttributes[MakeBoxes, {HoldAllComplete}];
        res
    ]

(* Erweiterung von Normal zur Extraktion des Polynoms aus QASeriesData *)
Normal[qasd:QASeriesData[___]] :=
    Plus @@ SeriesItems[qasd]

(* Erweiterung zur Darstellung von QASeriesData analog zu SeriesData *)
MakeBoxes[qasd:QASeriesData[vars_List, locs_List, factors_List, order_Integer], StandardForm] :=
    Module[ {res, si, rb, ps, vb},
        si = SeriesItems[qasd];
        si = Select[si, (# =!= 0 && # =!= QuantumAlgebra`Operator[0])&];
        rb = Map[MakeBoxesHelper, si];
        ps = Table[If[idx < Length[si] && !FreeQ[rb[[idx + 1]], List["-", __]], "", "+"], {idx, 1, Length[si]}];
        rb = {rb, ps} // Transpose // Flatten;
        vb = Map[MakeBoxesHelper, vars];
        vb = {vb, Table[",", Length[vars]]} // Transpose // Flatten;
        vb = Drop[vb, -1];
        rb = Append[rb, SuperscriptBox[RowBox[{"O", "[", Sequence @@ vb, "]"}], order + 1]];
        res = RowBox[rb];
        ClearAttributes[InterpretationBox, {HoldAllComplete}];
        SetAttributes[InterpretationBox, {HoldAll}];
        res = InterpretationBox[Evaluate[res], qasd];
        ClearAttributes[InterpretationBox, {HoldAll}];
        SetAttributes[InterpretationBox, {HoldAllComplete}];
        res
    ]

(* Nichtkommutative Multiplikation von QASeriesData *)
QASeriesData[vars1_List, locs1_List, factors1_List, order1_Integer] ** 
	QASeriesData[vars2_List, locs2_List, factors2_List, order2_Integer] :=
    Module[ 
    	{
    		combineVars, combineLocs, translateComb, result, vars, locs, factors, order,
    		noVars1 = Length[vars1], noVars2 = Length[vars2], noVars,
    		curOrder, curOrder1, curOrder2, combs1, comb1, fac1, combs2, comb2, fac2, tcomb
    	},
        combineVars[] :=
            Union[vars1, vars2];
        combineLocs[] :=
            Module[ {v, l},
                v = Join[vars1, vars2];
                l = Join[locs1, locs2];
                Map[l[[Position[v, #][[1, 1]]]]&, vars]
            ];
        translateComb[cb_Integer, v_List, rder_Integer] :=
            Module[ {nov = Length[v], id},
                id = IntegerDigits[cb, nov, rder] // Reverse;
                id = Map[(Position[vars, v[[# + 1]]][[1, 1]] - 1)&, id];
                FromDigits[id // Reverse, noVars]
            ];
        vars = combineVars[];
        noVars = Length[vars];
        locs = combineLocs[];
        order = Min[order1, order2];
        factors = Table[Table[QuantumAlgebra`Operator[0], noVars^i], {i, 0, order}];
        result = QASeriesData[vars, locs, factors, order];
        For[curOrder = 0, curOrder <= order, curOrder = curOrder + 1,
        	For[curOrder1 = 0, curOrder1 <= curOrder, curOrder1 = curOrder1 + 1,
        		curOrder2 = curOrder - curOrder1;
        		combs1 = noVars1^curOrder1;
        		combs2 = noVars2^curOrder2;
        		For[comb1 = 0, comb1 < combs1, comb1 = comb1 + 1,
        			fac1 = factors1[[curOrder1 + 1, comb1 + 1]];
        			For[comb2 = 0,comb2 < combs2, comb2 = comb2 + 1,
        				fac2 = factors2[[curOrder2 + 1, comb2 + 1]];
        				tcomb = translateComb[comb1, vars1, curOrder1] + translateComb[comb2, vars2, curOrder2] * noVars^curOrder1;
        				result[[3]][[curOrder + 1, tcomb + 1]] = (result[[3]][[curOrder + 1, tcomb + 1]] + fac1 ** fac2);
        			]
        		]
        	]
        ];
        result
    ]


(* Generiert eine Liste von Parametern mit Subscript *)
Params[param_Symbol, no_Integer] :=
    Table[Subscript[param, i], {i, no}]

(* Generiert eine Linear Form aus Parametern und Operatoren *)
LinearForm[gen_Symbol, param_Symbol, sgn_Integer, no_Integer] :=
	Module[{gens, params},
        gens = Table[QuantumAlgebra`Operator[gen, {i}, {}], {i, no}];
        params = Params[param, no];
        -I sgn params.gens
	]

(* Bildet den Gradient einer Linearform nach ihren Parametern *)	
GradLinearForm[lf_, params_List] := Table[QAD[lf, params[[i]]], {i, 1, Length[params]}]	

(* Von einer gegebenen Funktion werden die Reihenentwicklungskoeffizienten bestimmt *)
SeriesCoefficients[func_, order_Integer] :=
	Module[{x},
		Table[SeriesCoefficient[Series[func[x], {x, 0, order}], i], {i, 0, order}]
	]

(* Von einer Linearform wird der Gradient bestimmt und aus diesem alle moeglichen Kombinationen gegebener Ordnung *)
FactorCombinations[lf_, params_List, order_Integer] :=
	Module[{combine, grad},
		combine[gr_List, 0] := { QuantumAlgebra`Operator[1] };
		combine[gr_List, 1] := gr;
		combine[gr_List, n_Integer] := Outer[NonCommutativeMultiply, Sequence@@Table[gr, n]] // Flatten;
		 
		grad = GradLinearForm[lf, params];
		combine[grad, order]
	]

(* Ermittelt alle Ableitungskombinationen einer bestimmten Ordnung *)
DerivCombinations[op_, params_List, order_Integer] :=
	Module[{combine, grad},
		combine[gr_List, 0] := { QuantumAlgebra`Operator[1] };
		combine[gr_List, 1] := Map[QAD[op, params, #]&, gr];
		combine[gr_List, n_Integer] := Outer[QAD[op, params, #]&, Sequence@@Table[gr, n]] // Flatten;
		 
		grad = Range[Length[params]];
		combine[grad, order]
	]

(* Die Reihenentwicklung einer Funktion von einer Linearform wird bestimmt *)
SeriesFunctionLinearForm[func_, gen_Symbol, param_Symbol, sgn_Integer, no_Integer, order_Integer] :=
    Module[ {res, vars, locs, lf},
		vars = Params[param, no];
		locs = Table[0, no];
    	lf = LinearForm[gen, param, sgn, no];    	
        res = {
        vars,
        locs,
        Table[
        	SeriesCoefficients[func, order][[k + 1]] FactorCombinations[lf, vars, k], 
        	{k, 0, order}
        ],
        order};
        Apply[QASeriesData, res]
    ]

SeriesFunctionOperator[func_, op_?SimpleOperatorQ, param_Symbol, no_Integer, order_Integer] :=
	Module[{res, vars, locs},
		vars = Params[param, no];
		locs = Table[0, no];
        res = {
        vars,
        locs,
        Table[
        	SeriesCoefficients[func, order][[k + 1]] FactorCombinations[op, vars, k], 
        	{k, 0, order}
        ],
        order};
        Apply[QASeriesData, res]
	]

(* Generiert eine parametrisierte Lie - Gruppe *)
LieGroup[gen_Symbol, param_Symbol, sgn_Integer, no_Integer] :=
	Exp[LinearForm[gen, param, sgn, no]]

(* Generiert eine parametrisierte Lie - Gruppe und fuehrt eine Reihenentwicklung durch *)
SeriesLieGroup[gen_Symbol, param_Symbol, sgn_Integer, no_Integer, order_Integer] :=
	SeriesFunctionLinearForm[Exp, gen, param, sgn, no, order]


Protect[Release[protected]]

End[]

EndPackage[]
